# ------------------------------------------------------------------
# FixedFont - Write text in a fixed width. If we can't find a fixed
#    width font, simulate one by rendering each character.
#
# Copyright 2005 Ron Dippold
# ------------------------------------------------------------------


import pygame.font, pygame.sprite
from pygame.locals import *


class FixedFont:
	"""
	A FixedFont class handles one font at one size. It will try to render
	text at a fixed spacing - if this is a fixed size font this is fast
	and easy. Otherwise it will draw it one char at a time.
	"""


	# Helper func - generate character in cache if necessary, return image
	def __generate( self, c ):
		# Check if it's in the cache
		c_img = self.char_cache.get( c, None )
		if not c_img:
			# If not, create it and put it in there
			c_img = self.font.render( c, self.aalias, self.color )
			self.char_cache[c] = c_img
		return c_img
	

	def __init__( self, _fontname, _points, _aalias, _color, _bold=False, _italic=False ):
		"""
		_fontname = system font name (or names, in list) to match
		_points = point size
		_aalias = antialiased?
		_color = fg color (r,g,b,[a]) of text
		_bold, _italic = bold or italic?
		"""
		# Create the font
		self.font = pygame.font.SysFont( _fontname, _points, _bold, _italic )

		# render different characters in the font. If they're the same
		# size then we say it's a fixed size font.
		size = self.font.size( " " )
		self.fixed = True
		for c in ".O-":
			if self.font.size( c )!=size:
				self.fixed = False
				break

		self.aalias = _aalias
		self.color = _color

		# if it's fixed size, it's easy. Just save the size
		if self.fixed:
			self.charsize = size
			self.points = _points

		# and if it's not, precache the numbers, get maximum size
		else:
			# Our cache of characters - nothing loaded yet.
			self.char_cache = {}
			size = [ size[0], size[1] ]

			for c in "*O0123456789":
				c_img = self.__generate( c )
				size[0] = max( size[0], c_img.get_width() )
				size[1] = max( size[1], c_img.get_height() )
			self.charsize = size


	def render( self, _text ):
		"""
		return an image of fixed spacing rendered text
		"""

		# If it's a fixed font, that's easy, just render it
		if self.fixed:
			return self.font.render( _text, self.aalias, self.color )

		# Render a blank surface
		image = pygame.Surface( ( len(_text)*self.charsize[0], self.charsize[1] ) ).convert()
		image.fill( (0,0,0) )

		# For each character, verify that the character is in the cache, or 
		x = 0
		for c in _text:
			c_img = self.__generate(c)
			# the x+... is centering the letter in the middle of the cell
			image.blit( c_img, (x+(self.charsize[0]-c_img.get_width())/2, 0) )
			x += self.charsize[0]

		image.set_colorkey( image.get_at((0,0)), RLEACCEL )
		return image

	def size( self, _text ):
		"""
		return the size of the image that would be rendered
		"""
		if self.fixed:
			return self.font.size( _text )
		else:
			return ( len(_text)*self.charsize[0], self.charsize[1] )

