# -----------------------------------------------------------------------------
# Plane, Fighter, Copter - planes that can patrol.
#
# This does not include Bomber, which is just a disguised Missile
#
# Copyright 2005 Ron Dippold
# -----------------------------------------------------------------------------
import os
import random
import pygame.sprite, pygame.image
from   pygame.locals import *

import Util

class Plane( pygame.sprite.Sprite ):
	"""
	A Copter() and Fighter() are both a Plane() - they share the same idle
	behavior - just trundle back and forth.

	NOTE: all concrete classes must set
	imageset_array = [y][x] array of ImageSet for each direction - left,up = [0,0], right,down = [2,2]
	"""


	STATE_IDLE = 1
	
	# ------------------------------
	# Planes need to know the following global info.
	# all set with these planes.static.xxx() calls.
	# ------------------------------
	playrect = None			# Play rectangle
	patroltop = None		# Top of patrol area
	ship_pos = None			# position of ship
	flying_in = False		# if true, flying in from far left, triple speed!
	patrol_count = 0		# no of planes patrolling

	hitwave_callback = None # callback when plane hits a wave
	WaveHelper = None		# Helper to get specific height of wave

	#@staticmethod
	def set_playarea( _playrect, _patroltop ):
		Plane.playrect = _playrect
		Plane.patroltop = _patroltop
	# Pre python 2.4 syntax to make it static - sigh
	set_playarea = staticmethod( set_playarea )

	#@staticmethod
	def set_wave_info(  _WaveHelper,_hitwave_callback ):
		Plane.WaveHelper = _WaveHelper
		Plane.hitwave_callback = _hitwave_callback
	# Pre python 2.4 syntax to make it static - sigh
	set_wave_info = staticmethod( set_wave_info )

	#@staticmethod
	def set_ship_pos( _pos ):
		Plane.ship_pos = _pos
	# Pre python 2.4 syntax to make it static - sigh
	set_ship_pos = staticmethod( set_ship_pos )

	#@staticmethod
	def set_difficulty( _level, _game ):
		Plane.level_difficulty = _level
		Plane.game_difficulty = _game
	# Pre python 2.4 syntax to make it static - sigh
	set_difficulty = staticmethod( set_difficulty )

	# ------------------------------
	# all planes in idle mode use the base position (and an offset).
	# ------------------------------
	base_x, base_y = 0, 0		# base position
	base_speed = 20
	patrol_speed = 60
	base_newdir = None			# new direction this frame, update your images

	# Set a new direction for the base set - then all the concrete planes
	# can see this and change their own direction
	#@staticmethod
	def __set_base_newdir():
		Plane.base_newdir = Plane.moves[Plane.move_idx][1]
		Plane.speed = ( Plane.base_speed * Plane.base_newdir[0],
						Plane.base_speed * Plane.base_newdir[1] )
	# Pre python 2.4 syntax to make it static - sigh
	__set_base_newdir = staticmethod( __set_base_newdir )
	

	# Calculate the series of moves that the planes go through forever
	#@staticmethod
	def set_idlerect( _rect ):
		# Start at the top left, start heading for top right
		Plane.base_x, Plane.base_y = _rect.topleft

		Plane.moves = []
		# Set up moves for the plane - [ ( (dest),(direction) ), ... ]
		# Bah, set this up so we could do more complex moves, like some up
		# and down, but just left-right looked better
		y = _rect.top
		Plane.moves += [ ( (_rect.right,y), (1,0) ) ]  # move right
		Plane.moves += [ ( (_rect.left, y), (-1,0) ) ] # move left

		Plane.move_idx = 0
		Plane.__set_base_newdir()
	# Pre python 2.4 syntax to make it static - sigh
	set_idlerect = staticmethod( set_idlerect )

	#@staticmethod
	def flyin_from( _pos ):
		""" Start flying in from the far left """
		Plane.base_x = -_pos
		Plane.flying_in = True
		Plane.patrol_count = 0
	# Pre python 2.4 syntax to make it static - sigh
	flyin_from = staticmethod( flyin_from )

	#@staticmethod
	def update_basepos( seconds ):
		""" Update the idle position of the swarm. """

		if Plane.flying_in:
			speed = Plane.patrol_speed * 2 * seconds
			if Plane.base_x>60:
				Plane.flying_in = False
		else:
			speed = Plane.base_speed * seconds
		
		# Get destination, speed from current move. Don't need new move yet.
		( (destx,desty), (dirx,diry) )= Plane.moves[ Plane.move_idx ]
		new_move = False
		# either dirx or diry is non-zero, just do one
		if dirx:
			Plane.base_x += dirx * speed
			if (dirx>0 and Plane.base_x>=destx) or (dirx<0 and Plane.base_x<=destx ):
				Plane.base_x = destx      # clamp it
				new_move = True
		else:
			Plane.base_y += diry * speed
			if (diry>0 and Plane.base_y>=desty) or (diry<0 and Plane.base_y<=desty ):
				Plane.base_y = desty      # clamp it
				new_move = True

		if new_move:
			Plane.move_idx = ( Plane.move_idx + 1 ) % len( Plane.moves )
			Plane.__set_base_newdir()
		else:
			Plane.base_newdir = None
	# Pre python 2.4 syntax to make it static - sigh
	update_basepos = staticmethod( update_basepos )

	# Choose which set of _image_array to use based on direction
	def set_imageset( self, _dir ):
		self.imageset = self.imageset_array[ _dir[1]+1 ][ _dir[0]+1 ].shared_copy()

	# --------------------------
	# New Plane
	# --------------------------
	def __init__( self, _id,  _offset, _callback ):
		"""
		_id = (row,col) is our ID in the plane array
		_offset = our offset position from the base plane position
		_callback = routine to call when we're dead (or None)

		"""

		# Dummy static Plane, just bail
		if _id==None:
			return
			
		# Always init the sprite
		pygame.sprite.Sprite.__init__( self )

		self.id = _id
		self.offset = _offset
		self.callback = _callback

		# Just moving baaaack and foooorth
		self.state = Plane.STATE_IDLE

		# and choose the first one to start with
		self.set_imageset( Plane.base_newdir )

		# force position and image set
		self.update( 0 )

	#
	# Move it depending on how much time has passed - we only get here if
	# we're in idle state, otherwise the subclass overrides
	#

	def update( self, seconds ):
		# Update our position
		self.pos = [ Plane.base_x + self.offset[0], 
					 Plane.base_y + self.offset[1] ]

		# if base changed directions, then we change image
		if Plane.base_newdir:
			self.set_imageset( Plane.base_newdir )
			
		self.image, self.rect = self.imageset.newframe( seconds, self.pos )

	# Our overridden kill - to do the callback
	def kill( self ):
		if not self.is_idle():
			Plane.patrol_count -= 1
		
		# remove ourself from any active groups
		pygame.sprite.Sprite.kill( self )
		# If we have a callback, use that
		if self.callback:
			self.callback( self )

	def get_y( self ):
		return int(self.pos[1])

	# Generally, concrete class should override
	def score( self ):
		return 0

	# This should be sufficient for most
	def is_idle( self ):
		"Are we just idling back and forth?"
		return self.state == Plane.STATE_IDLE

	def my_idle_position( self ):
		"""return where I'd be idling if I was (and I might be!)"""
		return ( Plane.base_x + self.offset[0], Plane.base_y + self.offset[1] )

	def close_enough_to_idle( self ):
		"""
		Return true if we're within a few pixels of being back to where
		we should be.
		"""

		return ( abs(self.destx-self.x)<3 ) and ( abs(self.desty-self.y)<3 )

	def idle_now( self ):
		"""
		Return a patrolling plane to idle state NOW
		"""

		Plane.patrol_count -= 1
		self.state = Plane.STATE_IDLE
		self.set_imageset( Plane.moves[Plane.move_idx][1] )
		Plane.update( self, 0)


# -----------------------------------------------------------------------------
#
# class Copter
#
# -----------------------------------------------------------------------------


class Copter( Plane ):
	"""
	A nasty attack helicopter!

	"""

	# 3x3 arrays - [0][0] is heading nw, [1][0] is heading w, etc
	# Required by Plane() - and we can use it too!
	imageset_array = None

	# A Copter in patrol does the following ( from top to bottom )
	# STATE_MOVEOUT - move out a little (horizontal) 
	# STATE_DROP    - move down to patrol height
	# STATE_SIDEWAYS- move sideways a little, maybe vertical
	# STATE_HOVER1   - Wait a few seconds
	# STATE_ATTACK  - hold position and shoot
	# STATE_HOVER2   - Wait a few seconds
	#   if some attacks left, go back to STATE_SIDEWAYS
	# STATE_EXIT    -  otherwise, fly off screen
	# STATE_RETURN  - fly down to ready position

	( STATE_MOVEOUT, STATE_DROP, STATE_HOVER1, STATE_SIDEWAYS, STATE_ATTACK, STATE_HOVER2, STATE_EXIT, STATE_RETURN ) = \
	  range(10,18)

	# callback for firing a bullet
	bullet_callback = None

	#@staticmethod
	def preload():
		# load the right images
		right = Util.imageset_loader( "copter_r" )
		# flip them to the left
		left = right.flip( True, False )
		# load the center images
		center = Util.imageset_loader( "copter_c" )
		# image array is just left, center, right for all up/down
		Copter.imageset_array = [[ left, center, right ]] * 3

		# load hover sound
		Copter.sound = Util.sound_loader( "copter" )
		Copter.sound_list = {}
	# Pre python 2.4 syntax to make it static - sigh
	preload = staticmethod( preload )
	
	#@staticmethod
	def set_bullet_callback( _callback ):
		"""
		Set bullet firing callback - _callback takes 3 args: self, _pos and _speed.
		_pos is bullet position, _speed is speed normalized to 0-1 in both dimensions.
		"""
		Copter.bullet_callback = _callback
	# Pre python 2.4 syntax to make it static - sigh
	set_bullet_callback = staticmethod( set_bullet_callback )

	# --------------------------
	# New sprite
	# --------------------------
	def __init__( self, _id, _offset, _callback ):
		"""
		New Copter
		_id is for debugging
		_offset is our offset from the Plane.base position
		"""

		if not Copter.imageset_array:
			Copter.preload()

		Plane.__init__( self,
						_id				= _id,
						_offset			= _offset,
						_callback		= _callback )

	def score( self ):
		"how much we're worth dead."
		if self.is_idle():
			return 20
		else:
			return 50


	# ---------------------------
	# Hover sound stuff
	# ---------------------------
	def sound_start( self ):
		if not Copter.sound_list:
			Copter.sound.play(-1)
		Copter.sound_list[self] = True

	def sound_stop( self ):
		if Copter.sound_list.has_key( self ):
			del Copter.sound_list[self]
		if not Copter.sound_list:
			Copter.sound.fadeout(250)

	def kill( self ):
		if not self.is_idle():
			self.sound_stop()
		Plane.kill( self )

	# ---------------------------
	# All this is our patrol logic
	# ---------------------------


	def return_to_idle( self ):
		"""Return to idle at our own speed"""
		if not self.is_idle():
			self.next_state( Copter.STATE_EXIT )
		
	def idle_now( self ):
		"""need to stop the hover sound, so intercept this, then call parent"""
		self.sound_stop()
		Plane.idle_now( self )

	def patrol( self ):
		"""Start patroling if we're not just moving on screen"""
		if Plane.flying_in or self.state!=Plane.STATE_IDLE:
			return
		Plane.patrol_count += 1
		self.x, self.y = self.my_idle_position() 

		self.attacks_left = 1 + int( min(Plane.level_difficulty,8) * random.random())/2
		self.next_state( Copter.STATE_MOVEOUT, Util.sign(Plane.speed[0]) )

		self.sound_start()
		

	def fire_bullet( self ):
		""" Fire a bullet at the ship """
		# figure out where the ship is, and get a gross speed
		pos = [int(self.x), self.rect.bottom]
		speedx = Plane.ship_pos[0] - pos[0]
		# Copter can't fire behind itself
		if Util.sign(speedx)!=self.attack_dir[0]:
			speedx = 0
		speedy = Plane.ship_pos[1] - pos[1]

		# then normalize the speed to 0-1
		speedmax = float(max( abs(speedx), abs(speedy) ))
		if speedmax==0.0:
			speedmax=1.0
		speedx /= speedmax
		speedy /= speedmax

		# Fire!
		self.bullet_callback( pos, [speedx,speedy] )

	def update( self, _seconds ):
		# If we're idling, just let base class handle it
		if self.is_idle():
			return Plane.update( self, _seconds )

		# If we're attacking, occasionally fire a bullet
		if self.state == Copter.STATE_ATTACK:
			self.attack_time += _seconds
			if self.attack_time >= self.attack_interval:
				self.fire_bullet()
				self.attack_time -= self.attack_interval
				self.attack_count -= 1
				if self.attack_count<=0:
					self.next_state()

		# wait till time up
		elif self.state in ( Copter.STATE_HOVER1, Copter.STATE_HOVER2 ):
			self.hover_time -= _seconds
			if self.hover_time <= 0:
				self.next_state()

		# chase where we should be with where we are, till we get there
		elif self.state == Copter.STATE_RETURN:
			self.destx, self.desty = self.my_idle_position()

			# Are we done returning?
			if self.close_enough_to_idle():
				self.idle_now()
				return

			# should we be going in y or x direction?
			newdir = (0, Util.sign( (int(self.desty)-int(self.y))/3 ) )
			if newdir[1]==0:
				newdir = ( Util.sign( int(self.destx)-int(self.x) ), 0 )

			if self.dir != newdir:
				self.dir = newdir
				self.set_imageset( self.dir )

		else:
			if (self.dir[0]<0 and self.x<=self.destx) or \
				   (self.dir[0]>0 and self.x>=self.destx):
				self.x = self.destx
				self.dir = ( 0, self.dir[1] )
			if (self.dir[1]<0 and self.y<=self.desty) or \
				   (self.dir[1]>0 and self.y>=self.desty):
				self.y = self.desty
				self.dir = ( self.dir[0], 0 )
			if self.dir == (0,0):
				self.next_state()


		# figure out how fast we're going - based on difficulty
		delta = Plane.patrol_speed * _seconds
		if Plane.game_difficulty==2:
			delta *= 1.5
		elif Plane.game_difficulty==3:
			delta *= 2
		self.y += self.dir[1]*delta
		self.x += self.dir[0]*delta

		self.pos = ( self.x, self.y )
		self.image, self.rect = self.imageset.newframe( _seconds, self.pos )


	def next_state( self, _state=None, _dir=None ):

		# Which state next, if not given?
		if not _state:
			# if in STATE_HOVER2 we go out if no attacks left, otherwise
			# back to STATE_SIDEWAYS to attack again
			if self.state == Copter.STATE_HOVER2:
				self.attacks_left -= 1
				if self.attacks_left <= 0:
					_state = Copter.STATE_EXIT
				else:
					_state = Copter.STATE_SIDEWAYS
			elif self.state == Copter.STATE_RETURN:
				_state = Plane.STATE_IDLE
			else:
				_state = self.state + 1

		# assume we want to be where we are
		self.destx, self.desty = self.x, self.y

		# STATE_MOVEOUT - Move left/right a little before going down
		if _state == Copter.STATE_MOVEOUT:
			if not _dir: _dir=1
			self.destx = self.x + 40 * _dir

		# STATE_DROP - drop down to just above the waves
		elif _state == Copter.STATE_DROP:
			self.desty = random.randrange( Plane.patroltop, Plane.WaveHelper.min_top()-10 )

		# STATE_SIDEWAYS - move sideways a little
		elif _state == Copter.STATE_SIDEWAYS:
			self.destx = random.randrange(10,Plane.playrect.width-20)
			if random.random()<0.4:
				self.desty = random.randrange( Plane.patroltop, Plane.WaveHelper.min_top()-10 )

		# STATE_HOVER2 - hover a bit
		elif _state in (Copter.STATE_HOVER1, Copter.STATE_HOVER2):
			#multiplier = max( 2 - (Plane.level_difficulty/10.0), 0 )
			#self.hover_time = random.random() * multiplier
			self.hover_time = 2 * random.random()/Plane.game_difficulty

		# STATE_ATTACK - attack!
		elif _state == Copter.STATE_ATTACK:
			# draw us facing ship
			self.attack_dir = ( Util.sign( Plane.ship_pos[0]-self.x ), 0 )
			self.set_imageset( self.attack_dir )
			self.attack_count = random.randrange(4,8)
			self.attack_time = 0
			self.attack_interval = random.randrange(10,40)/100.0

		# STATE_EXIT - move off screen
		elif _state == Copter.STATE_EXIT:
			if self.x<Plane.playrect.centerx:
				self.destx = -self.rect.width
			else:
				self.destx = Plane.playrect.right + self.rect.width

		# STATE_RETURN - return to original position
		elif _state == Copter.STATE_RETURN:
			self.x = random.randrange(-20,20) + self.my_idle_position()[0]
			self.y = -20
			self.destx, self.desty = self.x, self.my_idle_position()[1]

		elif _state == Plane.STATE_IDLE:
			self.idle_now()
			return

		else:
			raise IndexError, "Unknown Copter state "+`_state`

		# Save the state
		self.state = _state

		# figure out what direction we're pointing
		self.dir = ( Util.sign(self.destx-self.x), Util.sign(self.desty-self.y) )

		if self.dir != (0,0):
			self.set_imageset( self.dir )
			


# -----------------------------------------------------------------------------
#
# class Fighter
#
# -----------------------------------------------------------------------------



class Fighter( Plane ):
	"""
	A loopy fighter plane

	"""


	# We want several color versions of the plane
	COLOR_LIST = ( (1,0,0), (0,1,0), (0,0,1), (1,0,1), (0,1,1), (1,1,1) ) # leave out (1,1,0)
	imageset_colors = None

	# A fighter on patrol loops down to the left or right. For left it's SW, S, SE, E.
	# for right it's SE, S, SW, W. After the last segment the plane can go south or north
	# in the same direction. So if it was going E, it might go SE, S, SW, W. Repeat until
	# we hit the waves or go offscreen.  May drop a bomb when going straight too.
	
    # SWLOOP NELOOP  NWLOOP SELOOP
	# /*      -\     /-     *\
    # |        |     |       |
	# \-      */     \*     -/
	#
	# Notice that the west loops end up going east and vice versa

	( STATE_SWLOOP, STATE_SELOOP, STATE_NWLOOP, STATE_NELOOP, STATE_RETURN ) = range( 20, 25 )

	LOOPS = {
		STATE_SWLOOP : ( (-1,1), (0,1), (1,1), (1,0) ),
		STATE_SELOOP : ( (1,1), (0,1), (-1,1), (-1,0) ),
		STATE_NWLOOP : ( (-1,-1), (0,-1), (1,-1), (1,0) ),
		STATE_NELOOP : ( (1,-1), (0,-1), (-1,-1), (-1,0) ) }

	# sin/cos of 45 degrees
	DIAGONAL = 0.7071


	# callback for dropping a bomb
	bomb_callback = None
	
	#@staticmethod
	def set_bomb_callback( _callback ):
		"""
		Set bomb dropping callback - _callback takes 2 args: self and _pos
		"""
		Fighter.bomb_callback = _callback
	# Pre python 2.4 syntax to make it static - sigh
	set_bomb_callback = staticmethod( set_bomb_callback )

	#@staticmethod
	def preload():
		# load the three images
		e = Util.imageset_loader( "fighter_r", _convert=False )
		n = Util.imageset_loader( "fighter_u", _convert=False )
		ne = Util.imageset_loader( "fighter_ur", _convert=False )
		# rotate (not flip!) to get se 
		se = ne.rotate( 90 )

		# Generate all the various color forms
		Fighter.imageset_colors = {}
		for i in range(len(Fighter.COLOR_LIST)):
			color = [ c*200 for c in Fighter.COLOR_LIST[i] ]

			# set the colors
			e.set_palette_at( 1, color )
			n.set_palette_at( 1, color )
			ne.set_palette_at( 1, color )
			se.set_palette_at( 1, color )

			# Get converted versions
			c_e = e.convert()
			c_n = n.convert()
			c_ne = ne.convert()
			c_se = se.convert()
				
			# The image array is just the e, n and ne images with a lot of flipping
			Fighter.imageset_colors[i] = [
				[ c_ne.flip( True, False ), c_n, c_ne ],
				[ c_e.flip( True, False ), None, c_e ],
				[ c_se.flip( True, False ), c_n.flip( False, True ), c_se ] ]

		Plane.sound = Util.sound_loader( 'fighter' )
	# Pre python 2.4 syntax to make it static - sigh
	preload = staticmethod( preload )

	# --------------------------
	# New Fighter
	# --------------------------
	def __init__( self, _id, _offset, _callback ):

		"""
		_id is (col,row) offset in planes table
		_offset is our offset from the Plane.base position
		_callback is for when we die
		"""

		# load the image and get the x center position
		if not Fighter.imageset_colors:
			Fighter.preload()

		# choose a color based on row
		color = _id[1] % len(Fighter.COLOR_LIST)
		self.imageset_array = Fighter.imageset_colors[color]
		
		Plane.__init__( self,
						_id				= _id,
						_offset			= _offset,
						_callback		= _callback )


	def score( self ):
		"how much we're worth dead."
		if self.state == Plane.STATE_IDLE:
			return 10
		else:
			return 20


	# ---------------------------
	# All this is our patrol logic
	# ---------------------------


	def drop_bomb( self ):
		""" Drop a bomb! """
		self.bomb_callback( [int(self.x), self.rect.bottom] )

	def return_to_idle( self ):
		"""Return to idle at our own speed"""
		if not self.is_idle():
			self.state = Fighter.STATE_RETURN

	def patrol( self ):
		"""Start patroling if we're not just moving on screen"""
		if Plane.flying_in or self.state!=Plane.STATE_IDLE:
			return

		Plane.patrol_count += 1
		self.sound.play()
		self.x, self.y = self.my_idle_position()

		# if we're heading east, then go SE. Otherwise SW
		if Plane.speed[0]>0:
			self.state = Fighter.STATE_SELOOP
		else:
			self.state = Fighter.STATE_SWLOOP
		self.loop_idx = 0

		# figure out next destination, orientation, etc.
		self.new_leg()


	def new_loop( self ):
		"""
		Choose a new loop - based on current loop.
			NW, SW -> SE or NE
			NE, SE -> SW or NW
		"""

		# Set up our choices based on the direction we were going - make it
		# much more likely to go down though
		if self.state in ( Fighter.STATE_NWLOOP, Fighter.STATE_SWLOOP ):
			if random.random()<0.8:
				self.state = Fighter.STATE_SELOOP
			else:
				self.state = Fighter.STATE_NELOOP
		else:
			if random.random()<0.8:
				self.state = Fighter.STATE_SWLOOP
			else:
				self.state = Fighter.STATE_NWLOOP

		# If the plane isn't below the patrol line yet, don't allow up
		if self.y<(Plane.patroltop+20):
			if self.state == Fighter.STATE_NWLOOP:
				self.state = Fighter.STATE_SWLOOP
			elif self.state == Fighter.STATE_NELOOP:
				self.state = Fighter.STATE_SELOOP

		# start at the first index of the new loop
		self.loop_idx = 0

		
	def new_leg( self ):
		"""
		Choose a new destination for the next leg of the loop
		"""

		# If we're at the end of the loop, we need to figure out a new one
		if self.loop_idx >= len(Fighter.LOOPS[self.state]):
			self.new_loop()

		# Get new direction, figure out a random length
		self.dir = Fighter.LOOPS[self.state][self.loop_idx]
		self.loop_idx += 1
		range = random.randrange( 10, 60 )

		# If it's diagonal, then do pythagorean on the length
		if self.dir[0] and self.dir[1]:
			range *= Fighter.DIAGONAL

		# new destination x,y
		self.destx = self.x + self.dir[0] * range

		# Try not to go off left edge even if it looks strange
		if self.dir[0]<0 and self.destx<5:
			range -= 5-self.destx
			self.destx = 5

		# Try not to go off right edge even if it looks strange
		if self.dir[0]>0:
			rightx = Plane.playrect.right - 5
			if self.destx>rightx:
				range -= self.destx-rightx
				self.destx = rightx

		# adjust the y - no checks here, we can hit waves
		self.desty = self.y + self.dir[1] * range

		# set image appropriately
		self.set_imageset( self.dir )

		if not self.dir[1] and random.random()<(.2*Plane.game_difficulty):
			self.drop_bomb()

		# print "New leg - pos:", self.x, self.y, " dest:", self.destx, self.desty, " dir:", self.dir


	def update( self, _seconds ):
		# If we're idling, just let base class handle it
		if self.is_idle():
			return Plane.update( self, _seconds )

		# Did we hit the waves?
		if self.rect.bottom >= Plane.WaveHelper.min_top():
			if self.rect.bottom >= Plane.WaveHelper.top( self.rect.centerx ):
				# We don't die, we just magically appear up above the play area
				# so we can come back down.
				self.state = Fighter.STATE_RETURN
				self.x = random.randrange(-20,20) + self.my_idle_position()[0]
				self.y = -60
				Plane.hitwave_callback( self,  self.rect.center )

		# If we're in return mode, try to get back to where we need to be,
		# by vertical till we're close, then just home in
		if self.state == Fighter.STATE_RETURN:
			self.destx, self.desty = self.my_idle_position()
			#print "Pos:", self.x, self.y, "  dest:", self.destx, self.desty

			# close enough - don't flipflop around all over when we get close
			if abs( self.desty-self.y )<3:
				self.desty = self.y
			if abs( self.destx-self.x )<3:
				self.destx = self.x
			
			if (self.desty-self.y) > 20:
				newdir = (0,1)
			elif (self.desty-self.y) < -20:
				newdir = (0,-1)
			else:
				newdir = ( Util.sign( int(self.destx)-int(self.x) ),
						  Util.sign( int(self.desty)-int(self.y) ) )
				
			# Are we there? Then just go back to idle mode
			if self.close_enough_to_idle():
				self.idle_now()
				return
			# Otherwise, set the new direction and image
			if self.dir != newdir:
				self.dir = newdir
				self.set_imageset( self.dir )

		self.x += self.dir[0]*Plane.patrol_speed * _seconds
		self.y += self.dir[1]*Plane.patrol_speed * _seconds
		self.pos = ( self.x, self.y )

		if self.state != Fighter.STATE_RETURN:
			# If we've reached our destination, choose a new destination
			if ( (self.dir[0]<0 and self.x<=self.destx) or (self.dir[0]>0 and self.x>=self.destx) or \
				 (self.dir[1]<0 and self.y<=self.desty) or (self.dir[1]>0 and self.y>=self.desty) ):
				self.x, self.y = self.destx, self.desty
				self.new_leg()


		# Finally update our positon and image frame
		self.image, self.rect = self.imageset.newframe( _seconds, self.pos )
